/** $lic$
 * Copyright (C) 2015-2016 by Massachusetts Institute of Technology
 *
 * This file is part of Whirltool.
 *
 * Whirltool is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * If you use this software in your research, we request that you reference
 * the Whirlpool paper ("Whirlpool: Improving Dynamic Cache Management with Static Data
 * Classification", Mukkara, Beckmann and Sanchez, ASPLOS-21, April 2016) as the
 * source in any publications that use this software, and that you send us a
 * citation of your work.
 *
 * Whirltool is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Whirltool.  If not, see <http://www.gnu.org/licenses/>.
 *
 **/
#ifndef HASH_H_
#define HASH_H_

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include "mtrand.h"

class H3HashFamily {
    private:
        uint32_t numFuncs;
        uint32_t resShift;
        uint64_t* hMatrix;

    public:
        H3HashFamily(uint32_t numFunctions, uint32_t outputBits, uint64_t randSeed = 123132127) : numFuncs(numFunctions) {
            MTRand rnd(randSeed);

            if (outputBits <= 8) {
                resShift = 3;
            } else if (outputBits <= 16) {
                resShift = 2;
            } else if (outputBits <= 32) {
                resShift = 1;
            } else if (outputBits <= 64) {
                resShift = 0;
            } else {
                printf("Hash function can't produce more than 64 bits of output!!");
                exit(1);
            }

            uint32_t words = 64 >> resShift;
            hMatrix = (uint64_t*)calloc(words*numFuncs, sizeof(uint64_t));
            for (uint32_t ii = 0; ii < numFuncs; ii++) {
                for (uint32_t jj = 0; jj < words; jj++) {
                    uint64_t val = 0;
                    for (int kk = 0; kk < 64; kk++) {
                        val = val << 1;
                        if (rnd.randInt() % 2 == 0) val++;
                    }
                    hMatrix[ii*words + jj] = val;
                }
            }
        }

        ~H3HashFamily() {};

        uint64_t hash(uint32_t id, uint64_t val) {
            uint64_t res = 0;
            assert(id >= 0 && id < numFuncs);

            // 8-way unrolled loop
            uint32_t maxBits = 64 >> resShift;
            for (uint32_t x = 0; x < maxBits; x+=8) {
                uint32_t base = (id << (6 - resShift)) + x;
                uint64_t res0 = val & hMatrix[base];
                uint64_t res1 = val & hMatrix[base+1];
                uint64_t res2 = val & hMatrix[base+2];
                uint64_t res3 = val & hMatrix[base+3];

                uint64_t res4 = val & hMatrix[base+4];
                uint64_t res5 = val & hMatrix[base+5];
                uint64_t res6 = val & hMatrix[base+6];
                uint64_t res7 = val & hMatrix[base+7];

                res ^= res0 ^ ((res1 << 1) | (res1 >> 63)) ^ ((res2 << 2) | (res2 >> 62)) ^ ((res3 << 3) | (res3 >> 61));
                res ^= ((res4 << 4) | (res4 >> 60)) ^ ((res5 << 5) | (res5 >> 59)) ^ ((res6 << 6) | (res6 >> 58)) ^ ((res7 << 7) | (res7 >> 57));
                res = (res << 8) | (res >> 56);
            }

            // Fold bits to match output
            switch (resShift) {
                case 0: //64-bit output
                    break;
                case 1: //32-bit output
                    res = (res >> 32) ^ res;
                    break;
                case 2: //16-bit output
                    res = (res >> 32) ^ res;
                    res = (res >> 16) ^ res;
                    break;
                case 3: //8-bit output
                    res = (res >> 32) ^ res;
                    res = (res >> 16) ^ res;
                    res = (res >> 8) ^ res;
                    break;
            }

            return res;
        }
};
#endif // HASH_H_
