/** $lic$
 * Copyright (C) 2015-2016 by Massachusetts Institute of Technology
 *
 * This file is part of Whirltool.
 *
 * Whirltool is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * If you use this software in your research, we request that you reference
 * the Whirlpool paper ("Whirlpool: Improving Dynamic Cache Management with Static Data
 * Classification", Mukkara, Beckmann and Sanchez, ASPLOS-21, April 2016) as the
 * source in any publications that use this software, and that you send us a
 * citation of your work.
 *
 * Whirltool is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Whirltool.  If not, see <http://www.gnu.org/licenses/>.
 *
 **/
#include <assert.h>
#include "cache.h"
#include <iomanip>
#include <iostream>
#include "llcProfiler.h"
#include "pin.H"
#include <stdio.h>

uint32_t logPageSize;

// Fast-forward handling vars //

static uint64_t icount = 0;
static uint64_t skipInstrCount = 0;
static uint64_t skipCount = 0;
static bool skipFlag = false;

uint64_t startIns;
uint64_t endIns;

static bool useHooks;
static bool inFastForward = true;

static uint64_t epochLength;

MissCurveProfiler* profiler;

// Cache analysis routine
void cacheAccess(uint64_t virtualAddr, bool load) {
    if (!skipFlag && !inFastForward) {
        // Here the virtual address is aligned to a word boundary
        virtualAddr = (virtualAddr >> 2) << 2;
        ((MissCurveProfiler*) profiler)->memReq(virtualAddr, load);
    }
}

#define ZSIM_MAGIC_OP_ROI_BEGIN         (1025)
#define ZSIM_MAGIC_OP_ROI_END           (1026)
#define ZSIM_MAGIC_OP_FFI_FF_EXIT       (1050)

#define ZSIM_MAGIC_OP_CREATE_SHARE      (2034)
#define ZSIM_MAGIC_OP_TAG_ADDRESS       (2035)
#define ZSIM_MAGIC_OP_UNTAG_ADDRESS     (2036)

#define ZSIM_MAGIC_OP_SKIP_INSTR_BEGIN  (2037)
#define ZSIM_MAGIC_OP_SKIP_INSTR_END    (2038)
#define ZSIM_MAGIC_OP_RECORD_ALLOCATION (2039)

// Forward declarations
ADDRINT HandleMagicOp(ADDRINT op, CONTEXT* ctxt);
void dumpProfilerStats();

// This function is called before every instruction is executed
VOID insCount() {
    if (!skipFlag) {
        if (useHooks) {
            if (!inFastForward) icount++;
        } else {
            icount++;
            if (icount >= startIns && inFastForward) {
                inFastForward = false;
                printf("FFI: Exiting fast-forward at %lu instructions\n", icount);
                return;
            } else if (icount >= endIns) {
                printf("FFI: Last ffi point reached. Exiting at %lu instructions\n", icount);
                dumpProfilerStats();
                exit(0);
            }
        }

        if ((((icount - startIns) % (epochLength)) == 0) && !inFastForward) {
            //printf("Resetting monitors at %lu instrs\n", icount);
            profiler->reset();
        }

        // print messages after every 1 billion instructions
        if (((icount % (1000*1000*1000)) == 0) && inFastForward) {
            printf("Some fast-forwarding going on\n");
        }
    } else {
        if (!inFastForward) skipInstrCount++;
    }
}

// Pin calls this function every time a new instruction is encountered
VOID Instruction(INS ins, VOID* v) {
    INS_InsertCall(ins, IPOINT_BEFORE, (AFUNPTR)insCount, IARG_END);

    // profile memory accesses
    if (INS_IsMemoryRead(ins))
        INS_InsertCall(ins, IPOINT_BEFORE, (AFUNPTR)cacheAccess,
                       IARG_MEMORYREAD_EA, IARG_END);
    if (INS_IsMemoryWrite(ins))
        INS_InsertCall(ins, IPOINT_BEFORE, (AFUNPTR)cacheAccess,
                       IARG_MEMORYWRITE_EA, IARG_END);

    if (INS_IsXchg(ins) && INS_OperandReg(ins, 0) == REG_RCX && INS_OperandReg(ins, 1) == REG_RCX) {
        INS_InsertCall(ins, IPOINT_BEFORE, (AFUNPTR) HandleMagicOp, IARG_REG_VALUE, REG_ECX, IARG_CONTEXT, IARG_RETURN_REGS, REG_EAX, IARG_END);
    }

}

ADDRINT HandleMagicOp(ADDRINT op, CONTEXT* ctxt) {
    switch (op) {
        uint32_t share;

        case ZSIM_MAGIC_OP_ROI_BEGIN:
            if (useHooks) {
                assert(inFastForward);
                inFastForward = false;
            }
            return 0;

        case ZSIM_MAGIC_OP_ROI_END:
           if (useHooks) {
               assert(!inFastForward);
               inFastForward = true;

               dumpProfilerStats();
               exit(0);
           }

        case ZSIM_MAGIC_OP_FFI_FF_EXIT:
            assert(inFastForward);
            inFastForward = false;

        case ZSIM_MAGIC_OP_SKIP_INSTR_BEGIN:
            if (!skipFlag) {
                skipFlag = true;
                skipCount++;
            }
            break;

        case ZSIM_MAGIC_OP_SKIP_INSTR_END:
            if (skipFlag) {
                skipFlag = false;
            }
            break;

        case ZSIM_MAGIC_OP_CREATE_SHARE:
            share = ((PageTable*) profiler)->createUserShare();
            return share;

        case ZSIM_MAGIC_OP_TAG_ADDRESS:
            ((PageTable*) profiler)->tagAddressUserShare(PIN_GetContextReg(ctxt, REG_RSI),
                        PIN_GetContextReg(ctxt, REG_RDX),
                        (uint32_t) PIN_GetContextReg(ctxt, REG_RBX));
            return 0;

        case ZSIM_MAGIC_OP_UNTAG_ADDRESS:
            ((PageTable*) profiler)->untagAddressUserShare(PIN_GetContextReg(ctxt, REG_RSI),
                        PIN_GetContextReg(ctxt, REG_RDX));
            return 0;
    }

    return 0;
}

// This function is called when the application exits
VOID Fini(INT32 code, VOID* v) {
    dumpProfilerStats();
}

// This knob will set the outfile name
KNOB<string> KnobOutputFile(KNOB_MODE_WRITEONCE, "pintool", "o", "results.out",
        "specify optional output file name");

void dumpProfilerStats() {
    FILE* outfile = NULL;
    assert(outfile = fopen(KnobOutputFile.Value().c_str(), "w"));
    profiler->dumpResults(outfile);
    fclose(outfile);
}

// Private cache parameters //

// This knob will set the param logPageSize
KNOB<uint32_t> KnobLogPageSize(KNOB_MODE_WRITEONCE, "pintool", "p", "12",
                             "specify the log of page size in bytes");

// This knob will set the cache param logNumRows
KNOB<uint32_t> KnobLogNumRows(KNOB_MODE_WRITEONCE, "pintool", "r", "8",
                            "specify the log of number of rows in the cache");

// This knob will set the cache param logBlockSize
KNOB<uint32_t> KnobLogBlockSize(
    KNOB_MODE_WRITEONCE, "pintool", "b", "6",
    "specify the log of block size of the cache in bytes");

// This knob will set the cache param associativity
KNOB<uint32_t> KnobAssociativity(KNOB_MODE_WRITEONCE, "pintool", "a", "8",
                               "specify the associativity of the cache");

// Fast-forward and ROI parameters //

KNOB<uint64_t> KnobROIStart(KNOB_MODE_WRITEONCE, "pintool", "start",
                                 "20000000000", "start of ROI(instructions)");

KNOB<uint64_t> KnobROILength(KNOB_MODE_WRITEONCE, "pintool", "length",
                                  "10000000000", "Length of ROI(instructions)");

KNOB<bool> KnobHooks(KNOB_MODE_WRITEONCE, "pintool", "hooks", "0",
                     "use roi hooks in application instead of instructions based sampling");

KNOB<uint64_t> KnobEpochLength(KNOB_MODE_WRITEONCE, "pintool",
            "el", "50000000", "epoch length(instructions)");

INT32 Usage()
{
    cerr << "This tool is used to profile number of accesses to each page in LLC cache" << endl;
    cerr << "Stats collection can be constrained to a specific region of interest(ROI)"
         << "specified in terms of instructions or through program-level hooks";
    cerr << endl << KNOB_BASE::StringKnobSummary() << endl;
    return -1;
}

// argc, argv are the entire command line, including pin -t <toolname> -- ...
int main(int argc, char* argv[]) {
    // Initialize pin
    if (PIN_Init(argc, argv)) { return Usage(); }

    logPageSize = KnobLogPageSize.Value();
    profiler = new MissCurveProfiler(KnobLogNumRows.Value(),
                                              KnobLogBlockSize.Value(),
                                              KnobAssociativity.Value(), logPageSize);

    startIns = KnobROIStart.Value();
    endIns = startIns + KnobROILength.Value();
    epochLength = KnobEpochLength.Value();

    useHooks = KnobHooks.Value();

    // Register Instruction to be called to instrument instructions
    INS_AddInstrumentFunction(Instruction, 0);

    // Register Fini to be called when the application exits
    PIN_AddFiniFunction(Fini, 0);

    // Start the program, never returns
    PIN_StartProgram();

    return 0;
}
