// Used to override malloc routines on systems that are using glibc.

#ifndef POOL_MALLOC_LIBC_OVERRIDE_GLIBC_H_
#define POOL_MALLOC_LIBC_OVERRIDE_GLIBC_H_

#include <features.h> // for __GLIBC__

#include "poolMalloc.h"

#ifndef __GLIBC__
# error libc_override_glibc.h is for glibc distributions only.
#endif

// In glibc, the memory-allocation methods are weak symbols, so we can
// just override them with our own.  If we're using gcc, we can use
// __attribute__((alias)) to do the overriding easily (exception:
// Mach-O, which doesn't support aliases).  Otherwise we have to use a
// function call.
#if !defined(__GNUC__) || defined(__MACH__)

// This also defines ReplaceSystemAlloc().
#include "libc_override_redefine.h"  // defines functions malloc()/etc

#else  // #if !defined(__GNUC__) || defined(__MACH__)

#include "libc_override_gcc_and_weak.h"

// We also have to do some glibc-specific overriding.  Some library
// routines on RedHat 9 allocate memory using malloc() and free it
// using __libc_free() (or vice-versa).  Since we provide our own
// implementations of malloc/free, we need to make sure that the
// __libc_XXX variants (defined as part of glibc) also point to the
// same implementations.  Since it only matters for redhat, we
// do it inside the gcc #ifdef, since redhat uses gcc.

#define ALIAS(pool_fn)   __attribute__ ((alias (#pool_fn)))
extern "C" {
  void* __libc_malloc(size_t size)                ALIAS(pool_malloc);
  void __libc_free(void* ptr)                     ALIAS(pool_free);
  void* __libc_realloc(void* ptr, size_t size)    ALIAS(pool_realloc);
  void* __libc_calloc(size_t n, size_t size)      ALIAS(pool_calloc);
  void __libc_cfree(void* ptr)                    ALIAS(pool_cfree);
  void* __libc_memalign(size_t align, size_t s)   ALIAS(pool_memalign);
  void* __libc_valloc(size_t size)                ALIAS(pool_valloc);
  void* __libc_pvalloc(size_t size)               ALIAS(pool_pvalloc);
  int __posix_memalign(void** r, size_t a, size_t s)  ALIAS(pool_posix_memalign);
}   // extern "C"
#undef ALIAS

#endif  // #if defined(__GNUC__) && !defined(__MACH__)


// We also have to hook libc malloc.  While our work with weak symbols
// should make sure libc malloc is never called in most situations, it
// can be worked around by shared libraries with the DEEPBIND
// environment variable set.  The below hooks libc to call our malloc
// routines even in that situation.  In other situations, this hook
// should never be called.
extern "C" {
static void* glibc_override_malloc(size_t size, const void *caller) {
  return pool_malloc(size);
}
static void* glibc_override_realloc(void *ptr, size_t size,
                                    const void *caller) {
  return pool_realloc(ptr, size);
}
static void glibc_override_free(void *ptr, const void *caller) {
  pool_free(ptr);
}
static void* glibc_override_memalign(size_t align, size_t size,
                                     const void *caller) {
  return pool_memalign(align, size);
}

// We should be using __malloc_initialize_hook here, like the #if 0
// code below.  (See http://swoolley.org/man.cgi/3/malloc_hook.)
// However, this causes weird linker errors with programs that link
// with -static, so instead we just assign the vars directly at
// static-constructor time.  That should serve the same effect of
// making sure the hooks are set before the first malloc call the
// program makes.
#if 0
#include <malloc.h>  // for __malloc_hook, etc.
void glibc_override_malloc_init_hook(void) {
  __malloc_hook = glibc_override_malloc;
  __realloc_hook = glibc_override_realloc;
  __free_hook = glibc_override_free;
  __memalign_hook = glibc_override_memalign;
}

void (* volatile __malloc_initialize_hook) (void) = &glibc_override_malloc_init_hook;
#endif

void* (* volatile __malloc_hook)(size_t, const void*)
    = &glibc_override_malloc;
void* (* volatile  __realloc_hook)(void*, size_t, const void*)
    = &glibc_override_realloc;
void (* volatile __free_hook)(void*, const void*)
    = &glibc_override_free;
void* (* volatile __memalign_hook)(size_t,size_t, const void*)
    = &glibc_override_memalign;

}   // extern "C"

// No need to write ReplaceSystemAlloc(); one of the #includes above
// did it for us.

#endif  // TCMALLOC_LIBC_OVERRIDE_GLIBC_INL_H_
