'''
/** $lic$
 * Copyright (C) 2015-2016 by Massachusetts Institute of Technology
 *
 * This file is part of Whirltool.
 *
 * Whirltool is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * If you use this software in your research, we request that you reference
 * the Whirlpool paper ("Whirlpool: Improving Dynamic Cache Management with Static Data
 * Classification", Mukkara, Beckmann and Sanchez, ASPLOS-21, April 2016) as the
 * source in any publications that use this software, and that you send us a
 * citation of your work.
 *
 * Whirltool is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Whirltool.  If not, see <http://www.gnu.org/licenses/>.
 *
 **/
'''

import numpy as np
from combineMissCurves import systemMissCurve, combinedMissCurve, convexifyCurve

def missCurveAreaDistance(P, Q):
    assert(P.shape == Q.shape)
    numIntervals = P.shape[0]
    intervalNorm = np.zeros(numIntervals)

    for time in range(numIntervals):
        Pc = P[time,:]
        Qc = Q[time,:]
        systemCurve = systemMissCurve(Pc, Qc)
        combinedCurve = combinedMissCurve(Pc, Qc)
        diff = combinedCurve - systemCurve
        intervalNorm[time] = np.sum(diff)
    norm = np.sum(intervalNorm)/numIntervals
    return norm

# Hierarchical clustering code adapted from
# http://www.maths.lth.se/matematiklth/personal/solem/downloads/hcluster.py
# by Jan Erik Solem
# Read http://www.janeriksolem.net/2009/04/hierarchical-clustering-in-python.html

def combineNode(v1,v2):
    assert(v1.shape == v2.shape)
    newMissCurves = np.zeros([v1.shape[0], v1.shape[1]], dtype=float)
    for time in range(v1.shape[0]):
        newMissCurves[time,:] = combinedMissCurve(v1[time,:], v2[time,:])
    return newMissCurves

def getPositiveId(id, n):
    if id >= 0:
        return id
    else:
        return -id + n - 1

def getChildren(id, n, Z):
    if id >= 0:
        return 1
    else:
        return Z[-id - 1, 3]

class cluster_node:
    # The indices of final missCurves being clustered in the original list of missCurves,
    # before filtering out low APKI miss curves. Used only for debug output in __str__()
    origIndices = None

    def __init__(self,vec,left=None,right=None,distance=0.0,id=None,count=1):
        self.left=left
        self.right=right
        self.id=id
        self.vec=vec
        self.distance=distance
        self.count=count # only used for weighted average

    def __str__(self):
        if self.origIndices is not None:
            return '[' + ' '.join([str(self.origIndices[x]) for x in self.leaves()]) + ']'
        else:
            return '[' + ' '.join([str(x) for x in self.leaves()]) + ']'

    def leaves(self):
        if self.left == None and self.right == None:
            return [self.id]

        l = []
        if self.left != None:
            l += self.left.leaves()
        if self.right != None:
            l += self.right.leaves()
        return l

# Main clustering routine
def hcluster(missCurves, distance=missCurveAreaDistance):
    distances = {} # distances is the cache of distance calculations
    currentclustid=-1

    # clusters are initially just the individual rows
    clust=[cluster_node(missCurves[i,:], id=i) for i in range(missCurves.shape[0])]

    n = missCurves.shape[0]
    Z = np.empty([n - 1, 4], dtype=float)
    iteration = 0
    while len(clust)>1:
        iteration += 1
        lowestpair=(0,1)

        closest=distance(clust[0].vec, clust[1].vec)
        print 'Comparing:', clust[0], clust[1], "Best:", closest

        # loop through every pair looking for the smallest distance
        for i in range(len(clust)):
            for j in range(i+1,len(clust)):
                if (clust[i].id,clust[j].id) not in distances:
                    distances[(clust[i].id,clust[j].id)] = distance(clust[i].vec, clust[j].vec)

                d=distances[(clust[i].id,clust[j].id)]
                print 'Comparing:', clust[i], clust[j], 'Distance:', distances[(clust[i].id, clust[j].id)], 'Best:', closest

                if d < closest:
                    closest=d
                    lowestpair=(i,j)

        newMissCurves = combineNode(clust[lowestpair[0]].vec, clust[lowestpair[1]].vec)

        Z[iteration - 1, 0] = getPositiveId(clust[lowestpair[0]].id, n)
        Z[iteration - 1, 1] = getPositiveId(clust[lowestpair[1]].id, n)
        Z[iteration - 1, 2] = closest
        Z[iteration - 1, 3] = getChildren(clust[lowestpair[0]].id, n, Z) +  getChildren(clust[lowestpair[1]].id, n, Z)

        # create the new cluster
        newcluster=cluster_node(newMissCurves,
                                left=clust[lowestpair[0]],
                                right=clust[lowestpair[1]],
                                distance=closest,id=currentclustid)
        print "Iteration" , iteration, "done ",
        print "Best", closest,"Lowest Pair:", clust[lowestpair[0]], clust[lowestpair[1]]

        # cluster ids that weren't in the original set are negative
        currentclustid-=1
        del clust[lowestpair[1]]
        del clust[lowestpair[0]]
        clust.append(newcluster)

    assert(iteration == n - 1)
    return Z

# Top-level routine that generates heirarchical clustering from input miss curves.
def wpLinkage(missCurves):
    missCurves = missCurves.swapaxes(0,1) # share * time * size

    convexCurves = np.zeros(missCurves.shape)

    for share in range(convexCurves.shape[0]):
        for time in range(convexCurves.shape[1]):
            convexCurves[share, time, :] = convexifyCurve(missCurves[share, time, :])

    print "Convexified curves"
    print "(shares, time samples, buckets) ->", convexCurves.shape
    return hcluster(convexCurves)
